/*
 * b. ems bus receiver example
 *
 * (C) 2009 by MK
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *
 * Protocol:
 * 
 *  0xAA 0x55           ( sync start )  
 *  EMS bus telegram    ( data       )
 *  data length         ( 1 byte     )
 *  0xAA 0x55           ( sync end   )
 */

#include <stdint.h>

#include <avr/io.h>
#include <avr/interrupt.h>

#define BLEN    200

static		    uint16_t buffer[BLEN]	    ;
static volatile uint16_t WriteIndex		    ;
static volatile uint16_t ReceiveIndex	    ;
static volatile uint8_t  BlinkCounter500ms  ;
static volatile uint8_t  Activity           ;

#define BAUD	9600
#define UBRR	( F_CPU / 16 / BAUD-1 )  // F_CPU is defined in the Makefile

#define avr_uart_DR			UDR
#define avr_uart_CSRB		UCSRB
#define avr_uart_CSRC		UCSRC
#define avr_uart_UBRRL		UBRRL
#define avr_uart_UBRRH		UBRRH


/*
#define avr_uart_DRE()	( UCSRA&_BV(UDRE ) )
#define avr_uart_RXC()	( UCSRA&_BV(RXC  ) )
*/
#define avr_uart_DRE()	bit_is_set( UCSRA, UDRE )
#define avr_uart_RXC()	bit_is_set( UCSRA, RXC  ) 


#define avr_uart_CSRB_INIT	( _BV(RXEN ) | _BV(TXEN ) | _BV(RXCIE) )
#define avr_uart_CSRC_INIT	( _BV(URSEL) | _BV(UCSZ0) | _BV(UCSZ1) )

// Pollin Eval Board V2.01 
#define LED_DATA_DDR    DDRD
#define LED_DATA_PORT   PORTD
#define LED_DATA_BIT    6 // led 2 

#define LED_TEST_DDR    DDRD
#define LED_TEST_PORT   PORTD
#define LED_TEST_BIT    5 // led 1

#define BTN_TEST_DDR    DDRD
#define BTN_TEST_PIN    PIND
#define BTN_TEST_BIT    2 // Button 1 

// timer for system state leds
#define TIMER1_INITVAL  0x10000 - ( F_CPU / 256 / 20 ); // 0.05 Hz blink (e.g. 2^16 - 16MHz / prescaler_256 / 20Hz )


void uart_init(void)
{
	uint8_t d       = avr_uart_DR	    ;
	avr_uart_CSRB	= 0				    ;
	avr_uart_CSRC	= 0				    ;
	avr_uart_UBRRH	= 0				    ;
	avr_uart_UBRRL	= 0				    ;
	d				= avr_uart_DR	    ;
	
	avr_uart_UBRRH	= (UBRR >> 8) & 0xFF;
	avr_uart_UBRRL	= (UBRR     ) & 0xFF;
	avr_uart_CSRC	= avr_uart_CSRC_INIT;
	avr_uart_CSRB	= avr_uart_CSRB_INIT;
}

void uart_write(uint8_t ch)
{
	while (!avr_uart_DRE())
	{
		// wait for UART
		__asm("nop");
	}
	avr_uart_DR = ch;
}

inline void uart_write_sync( void)
{
    uart_write( 0xAA );
	uart_write( 0x55 );
}

// interrupt handler 
ISR( SIG_UART_RECV )
{
	uint8_t  st	= 0;
	uint16_t x	= 0;

    cli();
	
	while ( 1 )
	{
		st = UCSRA;
		
		if ( !bit_is_set( st, RXC ) )
		{
			break;
		}
		
		// frame error flag -> end of telegram
		if ( bit_is_set( st, FE ) )
		{
            Activity = 1;
            LED_DATA_PORT ^= _BV( LED_DATA_BIT ); // toggle activity led 

			// mark highest bit as frame end  
			x = 0x8000;
		}
		else
		{
			// data
			x = 0x0000;
		}

		buffer[WriteIndex] = x | UDR;

		WriteIndex++;
		if ( WriteIndex == BLEN )
		{
			WriteIndex = 0;
		}
	}

    sei();
}
                    

// Timer ISR 50ms
ISR ( SIG_OVERFLOW1 ) 
{
    cli();

    TCNT1 = TIMER1_INITVAL; // 62411 for 0.05 Hz blink 

    // 500ms system blink 
    BlinkCounter500ms++;
    if( 10 == BlinkCounter500ms )
    {
        BlinkCounter500ms = 0;
        LED_TEST_PORT ^= _BV( LED_TEST_BIT ); // toggle run led 
    }


    /*
    // activity led 50ms blink 
    if( 0 != Activity )
    {
        LED_DATA_PORT ^= _BV( LED_DATA_BIT ); // toggle activity led 
    }
    else
    {
        // no communication -> port bit off 
        LED_DATA_PORT   &= ~_BV( LED_DATA_BIT );
    }
    */

    if( 0 == Activity )
    {
        // no communication -> port bit off 
        LED_DATA_PORT   &= ~_BV( LED_DATA_BIT );
    }
    
    if( 0 == BlinkCounter500ms )
    {
        Activity = 0;
    }
    
    sei();  
}


void init_timer (void) 
{  
    // 16MHz / 256 (prescaler) = 62500 steps/s
    // -> 0.05s == 3125 steps
    // -> 2^16 - 3125 = 62411 (initial value counter register TCNT1)

    BlinkCounter500ms   = 0;

    TIMSK    = _BV(TOIE1)       ; // Timer Overflow Interrupt enable 
    TCNT1    = TIMER1_INITVAL   ; // Initialize Timer
    TCCR1B   = _BV(CS12)        ; // Prescaler 256  
} 


int main()
{
	uint8_t ByteReceived	= 0;
	uint8_t DataLength		= 0;
	uint8_t SyncWithEmsBus	= 0;

    Activity                = 0;
	MCUSR			        = 0;
	
	// init watchdog
	ByteReceived = WDTCR;
	WDTCR = (1 < WDCE) | (1 < WDE);
	WDTCR = 0;

	// leds 
    LED_DATA_DDR    |=  _BV( LED_DATA_BIT );
    LED_TEST_DDR    |=  _BV( LED_TEST_BIT ); 
    BTN_TEST_DDR    &= ~_BV( BTN_TEST_BIT );
    
    
    LED_DATA_PORT   |= _BV( LED_DATA_BIT );
    
	
    init_timer();
	
	// init receive buffer
	ReceiveIndex = 0;	
	WriteIndex   = 0;

	
	uart_init();
	
	// sync start
	uart_write_sync();

    sei();
	
	while( 1 ) // endless main loop
	{
		cli();
		if ( ReceiveIndex != WriteIndex )
		{
			ByteReceived = 1;
		}
		sei();
		
		if ( ByteReceived )
		{
			if ( SyncWithEmsBus )
			{
				uart_write( buffer[ReceiveIndex] & 0xff );
				DataLength++;
			}
			
			// check for frame end
			if ( buffer[ ReceiveIndex ] > 0xFF )
			{
				// data length 
				uart_write( DataLength );
				
				// sync end 
				uart_write_sync();
				
				// sync start 
				uart_write_sync();
				
				// reset values 
				DataLength	    = 0;
				SyncWithEmsBus  = 1;
			}
			
            ReceiveIndex++;
			if ( BLEN == ReceiveIndex )
			{
				ReceiveIndex = 0;
			}
			
			ByteReceived = 0;
		}
		else
		{	
            // wait for interrupt 
			__asm("sleep");
             
		}
	}
    
	return 0;
}
